/****************************************************************************
 *                                                                          *
 *             Copyright 1992  Novell, Inc.  All Rights Reserved.           *
 *                                                                          *
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND        *
 * TREATIES.  THIS WORK MAY BE USED, COPIED, DISTRIBUTED, REVISED, MODIFIED *
 * TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED   *
 * RECAST, TRANSFORMED, OR ADAPTED  SO LONG AS THIS COPYRIGHT NOTICE IS		*
 * MAINTAINED.  ANY OTHER USE OR USE WITHOUT MAINTAINING THIS COPYRIGHT		*
 * COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 			*
 *                                                                          *
 ***************************************************************************/


/**/
/*                                                                       */
/*        Multi Protocol Time Server Client using TLI                    */
/*   		                                                                  */
/*   This is a time server client.  It works with multiple protocols.    */
/*   Specifically TCP and SPX.  After determining the service address    */
/*   it establishes connection with the time server using the correct	   */
/*   transport protocol.  Once a connection is established it will		     */
/*   make a time request, wait for the reply, display the time and then  */
/*   exit.								                                                       */
/*                                                                       */
/**/


/**** Include Files ****/

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stropts.h>
#include <tiuser.h>
#include <sys\types.h>
#include <sys\socket.h>
#include <netinet\in.h>
#include <time.h>
/***************************************************************************/


/**** Environment Configuration ****/

#ifndef TRUE
	#define TRUE	1
	#define	FALSE	0
#endif
#ifndef	nilp
	#define	nilp(s)		((s *)0)
#endif

#define	STACK_SIZE	(8 * 1024)
#define	nop

/* Protocol Control Array Indicies */
#define I_SPX				0
#define	I_TCP				1
#define	I_TP4				2
#define I_ADSP				3
#define PC_TABLE_SIZE		4		/* Protocol control array size */

#define LISTEN_Q_LENGTH		5		/* Max outstanding connect indications */

/* Protocol Support control flags */
#define PS_REGISTER			0x0001	/* Register service with directory */
#define PS_DEREGISTER		0x0002	/* Deregister service */
#define PS_LOCATE			0x0004	/* Locate a service */
#define	PS_ACTIVATE			0x0010	/* Service is once again active */
#define PS_DEACTIVATE		0x0020	/* Service is temporarily down */
#define PS_FAILURE			0x0040	/* Service has failed */
#define PS_LISTEN			0x0100	/* Establish listening address */
#define PS_CONNECT			0x0200	/* Establish client address */
#define PS_SERVICE			0x0400	/* Establish connection handler addr */
#define PS_BFORMAT			0x1000	/* Format tbind address into buffer */
#define PS_CFORMAT			0x2000	/* Format tcall address into buffer */
/***************************************************************************/


/**** General data types ****/

typedef struct t_bind	tbind_t;
typedef struct t_call	tcall_t;
typedef struct t_discon	tdiscon_t;
typedef struct t_info	tinfo_t;
typedef struct netbuf	netbuf_t;
typedef struct sockaddr_in
						ipaddr_t;
/***************************************************************************/


/**** Local data types ****/

typedef struct protocolControl {
	int			 pc_fd;				/* Listening fd */
									/* Protocol disconnect service func */
	void		(*pc_discon)(struct protocolControl *pc);
									/* Protocol address service function */
	void		*(*pc_addrSup)(int fd,
						tbind_t *tbind, tcall_t *tcall, int flags);
	char		*pc_protocol;		/* Protocol name for t_open */
	tinfo_t		*pc_tinfo;			/* Protocol info struct */
	tbind_t		*pc_tbind;			/* Server address struct */
	tcall_t		*pc_tcall;			/* Protocol connect control struct */
} pcontrol_t;
/***************************************************************************/


/**** Time Server controls ****/
/* Function Numbers */
#define TS_TIME_REQUEST		1			/* Function */
#define 	TS_TR_VER			1		/* Function Version */
#define TS_TIME_SYNC		2
#define 	TS_TY_VER			1
#define TS_TIME_SET			3
#define 	TS_TS_VER			1

/* Completion Codes */
#define TS_INVALID_REQUEST		-1
#define TS_UNSUPORTED_VERSION	-2
#define TS_REQUEST_COMPLETE	 	 0

/* Function structures */
typedef struct timeServerHeader {	/* Size should be multiple of DWORD */
	long int	function;
	long int	length;
	short int	version;
	short int	status;
} tsh_t;

typedef struct timeServerReply {
	tsh_t		header;
	time_t		time;
} tstr_t;
/***************************************************************************/


/**** Local Prototypes ****/

void 		 PCCleanup(pcontrol_t *pc);

pcontrol_t	*PCClientInit(char *protocol,
				 void (*discon)(int),
				 void *(*addrSup)(int, tbind_t *, tcall_t *, int),
				 int *error);

void		*SPXAddressSupport(int fd,
				tbind_t *tbind,
				tcall_t *tcall,
				int flags);

void 		 SPXDisconSupport(pcontrol_t *pc);

void		*TCPAddressSupport(int fd,
				tbind_t *tbind,
				tcall_t *tcall,
				int flags);

void 		 TCPDisconSupport(pcontrol_t *pc);

void		 TError(char *fmt, ...);

void		 TimeServ(pcontrol_t *pc);
/***************************************************************************/


/**** Local Variables ****/

/***************************************************************************/


/**** Main - TLI example multi-protocol client
 *
 * Input:	None
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	Exit Code				0 = Successful
 *									1 = Initialization Error
 *									2 = Protocol Error
 *									3 = Programming Error
 *
 * Comment:	Usage:  TIMECLI
 *
 ****/

void
main (void)
{
	int			 error,
				 exitStatus = 1,
				 i;
	pcontrol_t	*pc;

/* Initialize the protocol control structure & open correct protocol stack */
	error = 0;
	if ((pc = PCClientInit("/dev/nspx", SPXDisconSupport,
			SPXAddressSupport, &error)) != nilp(pcontrol_t))
		nop;
	else if ((pc = PCClientInit("/dev/tcp", TCPDisconSupport,
			TCPAddressSupport, &error)) != nilp(pcontrol_t))
		nop;

	/* Check other protocols here */

/* If Init fails because of no memory or no Time Server then exit */
	if (pc == nilp(pcontrol_t))
	{
		if (error == ENOMEM)
			printf("Out of Memory, unable to initialize; terminating.\n");
		else
			printf("Unable to locate a Time Server; terminating.\n");
		goto ErrorExit;
	}

/* Connect to the Time Server */
	if (t_connect(pc->pc_fd, pc->pc_tcall, pc->pc_tcall) == -1)
	{
		if (t_errno == TLOOK && t_look(pc->pc_fd) == T_DISCONNECT)
			(*pc->pc_discon)(pc);
		else
			TError("Unable to connect to Time Server at %s %s.",
				pc->pc_protocol,
				(*pc->pc_addrSup)(pc->pc_fd, pc->pc_tbind, pc->pc_tcall,
					PS_CFORMAT));
		goto ErrorExit;
	}
	TimeSync(pc);
	
/* Terminate the connection */
	if (t_snddis(pc->pc_fd, (struct t_call *)0) == -1)
	{
		TError("t_snddis failed");
		goto ErrorExit;
	}

/* Cleanup the protocol */
	exitStatus = 0;
	PCCleanup(pc);

ErrorExit:
	exit(exitStatus);
}
/***************************************************************************/


/**** PCCleanup - Cleanup and Free a Protocol Control Structure
 *
 * Input:	pcontrol_t	*pc;		protocol control structure to cleanup
 *
 * Output:	None
 *
 * Display:	None
 *
 * Returns:	None
 *
 * Comment:	This routine will cleanup an allocated pcontrol_t structure
 *			by dealocating all memory.
 ****/

void
PCCleanup(pcontrol_t *pc)
{
	t_free((char *)pc->pc_tcall, T_CALL_STR);
	t_free((char *)pc->pc_tbind, T_BIND_STR);
	t_unbind(pc->pc_fd);
	t_close(pc->pc_fd);
	t_free((char *)pc->pc_tinfo, T_INFO_STR);
	free(pc);
}
/***************************************************************************/


/**** PCClientInit - Initialize a Protocol Control Structure
 *
 * Input:	char		*protocol;	   protocol name
 *			void		(*discon)();   disconnect support routine
 *			void		*(*addrSup)(); service registration routine
 *
 * Output:	int			error;		   receives error value if return NULL
 *
 * Display:	Yes
 *
 * Returns:	pcontrol_t	*pc;		   Address of initialized struct
 *
 * Comment:	This routine will allocate and initialize a protocol control
 *			structure.  If any allocation fails or the specified protocol
 *			is not available then PCClientInit cleans up before returning
 *			NULL.  Otherwise it returns the protocol control structure.
 ****/

pcontrol_t
*PCClientInit(char *protocol,
			 void (*discon)(int),
			 void *(*addrSup)(int, tbind_t *, tcall_t *, int),
			 int *error)
{
	int			 fd;
	pcontrol_t	*pc;
	tinfo_t		 tinfo;

	if ((pc = (pcontrol_t *)malloc(sizeof(pcontrol_t))) == nilp(pcontrol_t))
	{
		*error = ENOMEM;
		goto ErrorExit0;
	}

	if ((pc->pc_fd = fd = t_open(protocol, O_RDWR, &tinfo)) == -1)
	{
		TError("t_open for transport '%s' failed", protocol);
		goto ErrorExit1;
	}
	if ((pc->pc_tinfo = (tinfo_t *)t_alloc(fd, T_INFO_STR, T_ALL)) ==
			nilp(tinfo_t))
	{
		*error = ENOMEM;
		goto ErrorExit2;
	}
	*pc->pc_tinfo = tinfo;
	if ((pc->pc_tcall = (tcall_t *)t_alloc(fd, T_CALL_STR, T_ALL)) ==
			nilp(tcall_t))
	{
		*error = ENOMEM;
		goto ErrorExit3;
	}
	pc->pc_tcall->addr.len = pc->pc_tcall->addr.maxlen = pc->pc_tinfo->addr;

/* Allocate Initialize t_bind structure */
	if ((pc->pc_tbind = (tbind_t *)t_alloc(fd, T_BIND_STR, T_ALL)) ==
			nilp(tbind_t))
	{
		*error = ENOMEM;
		goto ErrorExit4;
	}
	pc->pc_tbind->addr.len = pc->pc_tbind->addr.maxlen = pc->pc_tinfo->addr;

/* Use protocol specific support routine to locate service & do t_bind */
	if (!(*addrSup)(fd, pc->pc_tbind, pc->pc_tcall, PS_LOCATE | PS_CONNECT))
		goto ErrorExit5;

	pc->pc_protocol = protocol;
	pc->pc_discon = discon;
	pc->pc_addrSup = addrSup;

	return pc;


ErrorExit5:
	t_free((char *)pc->pc_tbind, T_BIND_STR);
ErrorExit4:
	t_free((char *)pc->pc_tcall, T_CALL_STR);
ErrorExit3:
	t_free((char *)pc->pc_tinfo, T_INFO_STR);
ErrorExit2:
	t_close(fd);
ErrorExit1:
	free(pc);
ErrorExit0:
	return nilp(pcontrol_t);
}
/***************************************************************************/


/**** SPX Address Parsing routines
 *
 * Input:	char		*addressBuffer;
 *
 * Output:	IPX_ADDR	*parsedAddress;
 *
 *
 * Display:	None
 *
 * Returns:	int			 result;		TRUE if addressBuffer was valid
 *
 * Comment:	Address must be in nnnnnnnn:aaaaaaaaaaaa;ssss notation with
 *			leading zeros optional.  nnnnnnnn = Network number,
 *			aaaaaaaaaaaa = Node number, ssss = Socket number.
 ****/

static int
InsertAddress(char *string, char *buf, int hexBytes)
{
	int			 i,
				 j,
				 value;
	char		 c;

	for (i = 0; i < hexBytes; i++)
	{
	/* build a byte from two nibbles */
		value = 0;
		for (j = 0; j < 2; j++)
		{
			value <<= 4;
			if ((c = toupper(*string)) >= '0' && c <= '9')
				value += c - '0';
			else if (c >= 'A' && c <= 'F')
				value += c - 'A' + 10;
			else
				return FALSE;
			string++;
		}
		*buf++ = value;
	}
	return TRUE;
}
/***************************************************************************/

static int
SPXParseAddress(char *text, IPX_ADDR *spxAddr)
{
	char		 temp[13],
				*cp;
	int			 len;

/* Does the number contain a net number? */
	if ((cp = strchr(text, ':')) == NULL)
		return FALSE;

/* Parse out the net number */
	*cp++ = '\0';
	if ((len = strlen(text)) < 1 || len > 8)
		return FALSE;
	strcpy(temp, "00000000");
	strcpy(&temp[8 - strlen(text)], text);
	if (!InsertAddress(temp, spxAddr->ipxa_net, 4))
		return FALSE;
/* Skip over the Network Number portion */
	text = cp;

/* Does the number contain a Socket number? */
	if ((cp = strchr(text, ';')) == NULL)
		return FALSE;

/* Parse out the socket number */
	*cp++ = '\0';
	if ((len = strlen(text)) < 1 || len > 4)
		return FALSE;
	strcpy(temp, "0000");
	strcpy(&temp[4 - strlen(cp)], cp);
	if (!InsertAddress(temp, spxAddr->ipxa_socket, 2))
		return FALSE;

/* Parse out the node number */
	if ((len = strlen(text)) < 1 || len > 12)
		return FALSE;
	strcpy(temp, "000000000000");
	strcpy(&temp[12 - strlen(text)], text);
	if (!InsertAddress(temp, spxAddr->ipxa_node, 6))
		return FALSE;
	
	return TRUE;
}
/***************************************************************************/


/**** SPXAddressSupport - Protocol specific address/service location support
 *
 * Input:	int 		 fd;		handle of stream disconnected
 *			tbind_t		*tbind;		bind struct to initialize
 *			tcall_t		*tcall;		call struct to initialize
 *			int			 flags;		registration control flags
 *
 * Output:	tbind_t		*tbind;		bind struct containing client address
 *			tcall_t		*tcall;		call struct containing service address
 *
 * Display:	Yes
 *
 * Returns:	void		*ptr;		tbind, NULL or static buf
 *
 * Comment:	This function is responsible for initializing the address for
 *			specified endpoint.  It can locate a service through a
 *			name service such as NetWare Directory Service or could use some
 *			other means of querying the network for the service and
 *			obtaining its address.  This is currently using a very simple
 *			method, display the information and let the user decide what
 *			to do.
 ****/

void
*SPXAddressSupport(int fd, tbind_t *tbind, tcall_t *tcall, int flags)
{
/* Process format address request */
	if (flags & (PS_BFORMAT | PS_CFORMAT))
	{
		static char	addrbuf[30];
		IPX_ADDR	*spxAddr;

		spxAddr = (IPX_ADDR *)((flags & PS_BFORMAT) ? (tbind->addr.buf) :
			tcall->addr.buf);
		sprintf(addrbuf,
			"%02x%02x%02x%02x:%02x%02x%02x%02x%02x%02x;%02x%02x",
			spxAddr->ipxa_net[0] & 0xFF, spxAddr->ipxa_net[1] & 0xFF,
			spxAddr->ipxa_net[2] & 0xFF, spxAddr->ipxa_net[3] & 0xFF,
			spxAddr->ipxa_node[0] & 0xFF, spxAddr->ipxa_node[1] & 0xFF,
			spxAddr->ipxa_node[2] & 0xFF, spxAddr->ipxa_node[3] & 0xFF,
			spxAddr->ipxa_node[4] & 0xFF, spxAddr->ipxa_node[5] & 0xFF,
			spxAddr->ipxa_socket[0] & 0xFF, spxAddr->ipxa_socket[1] & 0xFF);
		return addrbuf;
	}

/* Process Name service lookup request */
	if (flags & PS_LOCATE)
	{
		char	inputBuffer[80];
		int		result;

		do
		{
			printf("Enter Time Server's SPX Address "
				   "in nnnnnnnn:aaaaaaaaaaaa;ssss form,\n"
				   "\tor enter 0 to skip:  ");
			gets(inputBuffer);
			if (strlen(inputBuffer) == 1 && inputBuffer[0] == '0')
			{
				printf("SPX based Time Server not requested.\n");
				return nilp(tbind_t);
			}
			if (!(result = SPXParseAddress(inputBuffer,
					(IPX_ADDR *)tcall->addr.buf)))
				printf("Invalid SPX Address, check address format.\n");
		}
		while (!result);
	}

/* Process Service connect bind request */
	if (flags & PS_CONNECT)
	{
	/* Allocate a dynamic SPX socket */
		if (t_bind(fd, nilp(tbind_t), tbind) == -1)
		{
			TError("t_bind in SPX address support routine failed");
			return nilp(tbind_t);
		}
	}

	if (flags & PS_LOCATE)
		return tcall;
	else
		return tbind;
}
/***************************************************************************/


/**** SPXDisconSupport - determine disconnect reason code and print message
 *
 * Input:	pcontrol_t	*pc;		control structure
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	Determine why the connection was terminated, by checking the
 *			reason code returned from t_rcvdis.  These are SPX specific
 *			codes and a message is printed accordingly.
 ****/

void
SPXDisconSupport(pcontrol_t *pc)
{
	tdiscon_t	 discon;

	if (t_rcvdis(pc->pc_fd, &discon) == -1)
	{
		t_error("t_rcvdis failed");
		ExitThread(0, 2);
	}
	switch(discon.reason)
	{
	case TLI_SPX_CONNECTION_FAILED:
		printf("Connection with %s %s terminated because\n"
			"\tthe connection failed.\n", pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;

	case TLI_SPX_CONNECTION_TERMINATED:
		printf("Connection with %s %s terminated by client.\n",
			pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;

	case TLI_SPX_MALFORMED_PACKET:
		printf("Connection with %s %s terminated because\n"
			"\tof internal SPX error.\n", pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;

	default:
		printf("Connection with %s %s terminated for unknown reason.\n",
			pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
	}
}
/***************************************************************************/


/**** TCP Address Parsing routines
 *
 * Input:	char		*addressBuffer;
 *
 * Output:	ipaddr_t	*parsedAddress;
 *
 *
 * Display:	None
 *
 * Returns:	int			 result;		TRUE if addressBuffer was valid
 *
 * Comment:	Address must be in nnn.nnn.nnn.nnn;pppp.  Where nnn.nnn.nnn.nnn
 *			is standard internet notation and pppp is the port number in hex.
 ****/

static int
TCPParseAddress(char *text, ipaddr_t *ipAddr)
{
	char		 temp[13],
				*cp;
	int			 i,
				 len;

	for (i = 0; i < 4; i++)
	{
	/* Does the number contain a divider? */
		if ((cp = strchr(text, i != 3 ? '.' : ';')) == NULL)
			return FALSE;

	/* Parse out next byte */
		*cp++ = '\0';
		if ((len = strlen(text)) < 1 || len > 3)
			return FALSE;
		((char *)&ipAddr->sin_addr.s_addr)[i] = atoi(text);
		text = cp;
	}

/* Parse out the port number */
	if ((len = strlen(text)) < 1 || len > 4)
		return FALSE;
	strcpy(temp, "0000");
	strcpy(&temp[12 - strlen(text)], text);
	if (!InsertAddress(temp, (char *)&ipAddr->sin_port, 2))
		return FALSE;
	
	return TRUE;
}
/***************************************************************************/


/**** TCPAddressSupport - Protocol specific address/service location support
 *
 * Input:	int 		 fd;		handle of stream disconnected
 *			tbind_t		*tbind;		bind struct to initialize
 *			tcall_t		*tcall;		call struct to initialize
 *			int			 flags;		registration control flags
 *
 * Output:	tbind_t		*tbind;		bind struct containing client address
 *			tcall_t		*tcall;		call struct containing service address
 *
 * Display:	Yes
 *
 * Returns:	void		*ptr;		tbind, NULL or static buf
 *
 * Comment:	This function is responsible for initializing the address for
 *			specified endpoint.  It can locate a service through a
 *			nameservice such as NetWare Directory Service or could use some
 *			other means of querying the network for the service and
 *			obtaining its address.  This is currently using a very simple
 *			method, display the information and let the user decide what
 *			to do.
 ****/

void
*TCPAddressSupport(int fd, tbind_t *tbind, tcall_t *tcall, int flags)
{
		ipaddr_t	*ipAddr;

/* Process format address request */
	if (flags & (PS_BFORMAT | PS_CFORMAT))
	{
		static char	addrbuf[30];

		ipAddr = (ipaddr_t *)((flags & PS_BFORMAT) ? (tbind->addr.buf) :
			tcall->addr.buf);
		sprintf(addrbuf, "%d.%d.%d.%d, port %x",
			((char *)&ipAddr->sin_addr.s_addr)[0],
			((char *)&ipAddr->sin_addr.s_addr)[1],
			((char *)&ipAddr->sin_addr.s_addr)[2],
			((char *)&ipAddr->sin_addr.s_addr)[3],
			htons(ipAddr->sin_port));
	}

/* Process Name service lookup request */
	if (flags & PS_LOCATE)
	{
		char	inputBuffer[80];
		int		result;

		do
		{
			printf("Enter Time Server's TCP Address "
				   "in nnn.nnn.nnn.nnn form,\n\tor enter 0 to skip:  ");
			gets(inputBuffer);
			if (strlen(inputBuffer) == 1 && inputBuffer[0] == '0')
			{
				printf("TCP based Time Server not requested.\n");
				return nilp(tbind_t);
			}
			if (!(result = TCPParseAddress(inputBuffer,
					(ipaddr_t *)tcall->addr.buf)))
				printf("Invalid TCP Address, check address format.\n");
		}
		while (!result);
	}

/* Process Service connect bind request */
	if (flags & PS_CONNECT)
	{
	/* Initialize everything to 0, then set required fields */
		ipAddr = (ipaddr_t *)tbind->addr.buf;
		memset(ipAddr, 0, tbind->addr.len);
		ipAddr->sin_family = AF_INET;

	/* Allocate a dynamic TCP port */
		if (t_bind(fd, tbind, tbind) == -1)
		{
			TError("t_bind in TCP address support routine failed");
			return FALSE;
		}
	}

	if (flags & PS_LOCATE)
		return tcall;
	else
		return tbind;
}
/***************************************************************************/


/**** TCPDisconSupport - determine disconnect reason code and print message
 *
 * Input:	pcontrol_t	*pc;		control structure
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	Determine why the connection was terminated, by checking the
 *			reason code returned from t_rcvdis.  These are TCP specific
 *			codes and a message is printed accordingly.
 ****/

void
TCPDisconSupport(pcontrol_t *pc)
{
	tdiscon_t	 discon;

	if (t_rcvdis(pc->pc_fd, &discon) == -1)
	{
		t_error("t_rcvdis failed");
		ExitThread(0, 2);
	}
	switch(discon.reason)
	{
	case ETIMEDOUT:
		printf("Connection with %s %s terminated because\n"
			"\tthe connection failed.\n", pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;

	case ECONNRESET:
		printf("Connection with %s %s terminated by client.\n",
			pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;

	case ENETDOWN:
		printf("Connection with %s %s terminated because\n"
			"\tTCP is either unloading or has been unloaded.\n",
			pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;

	default:
		printf("Connection with %s %s terminated for unknown reason.\n",
			pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind,
			pc->pc_tcall, PS_BFORMAT));
		break;
	}
}
/***************************************************************************/


/**** TError - t_error with replacable parameters
 *
 * Input:	char		*fmt;		display string
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	t_error does not provide for replacable parameters.  This
 *			routine front ends t_error to provide this functionality.
 *			routine susceptable to local buffer overflow problems.
 *
 ****/

void
TError(char *fmt, ...)
{
	va_list			argptr;
	char			buf[512];

	va_start(argptr, fmt);
	vsprintf(buf, fmt, argptr);
	va_end(argptr);
	t_error(buf);
}
/***************************************************************************/


/**** TimeSync - Request a time update and wait for reply
 *
 * Input:	pcontrol_t	*pc;		control struct for this connection
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	This is the send/receive thread for each connection.  It is
 *			started by the BeginThread call in ListenProcessor.  If any
 *			error occurs on any TLI call it terminates the thread.
 *			NOTE:  The parameter "pc" is released.
 ****/

void
TimeSync(pcontrol_t *pc)
{
	int			 flags,
				 len;
	tsh_t		 request;
	tstr_t		 reply;
	
	flags = 0;			/* Initialize for t_rcv call */

/* Initialize the request and send it */
	request.function = TS_TIME_REQUEST;
	request.length = sizeof(tsh_t);
	request.version = TS_TR_VER;
	request.status = 0;
	if (t_snd(pc->pc_fd, (char *)&request, request.length, 0) == -1)
	{
		TError("Sending the request failed.");
		return;
	}
	flags = 0;

/* Wait for the reply */
	if ((len = t_rcv(pc->pc_fd, (char *)&reply, sizeof(reply), &flags)) == -1 ||
			len < sizeof(tstr_t))
	{
		t_error("Receiving the reply failed");
		if (t_errno == TLOOK && t_look(pc->pc_fd) == T_DISCONNECT)
			(*pc->pc_discon)(pc);
		return;
	}
	if (reply.header.status == TS_REQUEST_COMPLETE)
		printf("Time server returned %s\n",asctime(localtime(&reply.time)));
	else
		printf("Time server returned ERROR %d\n", reply.header.status);
	return;		
}
/***************************************************************************/
/***************************************************************************/

