/****************************************************************************
 *                                                                          *
 *             Copyright 1992  Novell, Inc.  All Rights Reserved.           *
 *                                                                          *
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND        *
 * TREATIES.  THIS WORK MAY BE USED, COPIED, DISTRIBUTED, REVISED, MODIFIED *
 * TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED   *
 * RECAST, TRANSFORMED, OR ADAPTED  SO LONG AS THIS COPYRIGHT NOTICE IS		*
 * MAINTAINED.  ANY OTHER USE OR USE WITHOUT MAINTAINING THIS COPYRIGHT		*
 * COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 			*
 *                                                                          *
 ***************************************************************************/


/**/
/*                                                                       */
/*        Multi Protocol Time Server using TLI                           */
/*   		                                                                  */
/*   This is a time server.  It works with multiple protocols.           */
/*   Specifically TCP and SPX.  After establishing the service address   */
/*   it registers the address and service type with the appropriate	     */
/*   naming service and then waits for connections from the clients	     */
/*   of any of the supported protocols.  Once a connection is 		      	  */
/*   established it will receive a service request and reply to that	    */
/*   request appropriately.  To terminate the service simply press	the	  */
/*   escape key.  All current connections will be completed and then	    */
/*   the service will exit.			                                           */
/*                                                                       */
/**/


/**** Include Files ****/

#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <process.h>
#include <errno.h>
#include <fcntl.h>
#include <stropts.h>
#include <tiuser.h>
#include <sys\types.h>
#include <sys\socket.h>
#include <netinet\in.h>
#include <time.h>
/***************************************************************************/


/**** Environment Configuration ****/

#ifndef TRUE
	#define TRUE	1
	#define	FALSE	0
#endif
#ifndef	nilp
	#define	nilp(s)		((s *)0)
#endif

#define	STACK_SIZE	(8 * 1024)
#define	nop

/* Protocol Control Array Indicies */
#define I_SPX				0
#define	I_TCP				1
#define	I_TP4				2
#define I_ADSP				3
#define PC_TABLE_SIZE		4		/* Protocol control array size */

#define LISTEN_Q_LENGTH		5		/* Max outstanding connect indications */

/* Protocol Support control flags */
#define PS_REGISTER			0x0001	/* Register service with directory */
#define PS_DEREGISTER		0x0002	/* Deregister service */
#define PS_LOCATE			0x0004	/* Locate a service */
#define	PS_ACTIVATE			0x0010	/* Service is once again active */
#define PS_DEACTIVATE		0x0020	/* Service is temporarily down */
#define PS_FAILURE			0x0040	/* Service has failed */
#define PS_LISTEN			0x0100	/* Establish listening address */
#define PS_CONNECT			0x0200	/* Establish client address */
#define PS_SERVICE			0x0400	/* Establish connection handler addr */
#define PS_BFORMAT			0x1000	/* Format tbind address into buffer */
#define PS_CFORMAT			0x2000	/* Format tcall address into buffer */
/***************************************************************************/


/**** General data types ****/

typedef struct t_bind	tbind_t;
typedef struct t_call	tcall_t;
typedef struct t_discon	tdiscon_t;
typedef struct t_info	tinfo_t;
typedef struct netbuf	netbuf_t;
typedef struct sockaddr_in
						ipaddr_t;
/***************************************************************************/


/**** Local data types ****/

typedef struct protocolControl {
	int			 pc_fd;				/* Listening fd */
									/* Protocol disconnect service func */
	void		(*pc_discon)(struct protocolControl *pc);
									/* Protocol address service function */
	void		*(*pc_addrSup)(int fd, tbind_t *tbind, int flags);
	char		*pc_protocol;		/* Protocol name for t_open */
	tinfo_t		*pc_tinfo;			/* Protocol info struct */
	tbind_t		*pc_tbind;			/* Server address struct */
	tcall_t		*pc_tcall;			/* Protocol connect control struct */
} pcontrol_t;
/***************************************************************************/


/**** Time Server controls ****/
/* Function Numbers */
#define TS_TIME_REQUEST		1			/* Function */
#define 	TS_TR_VER			1		/* Function Version */
#define TS_TIME_SYNC		2
#define 	TS_TY_VER			1
#define TS_TIME_SET			3
#define 	TS_TS_VER			1

/* Completion Codes */
#define TS_INVALID_REQUEST		-1
#define TS_UNSUPORTED_VERSION	-2
#define TS_REQUEST_COMPLETE	 	 0

/* Function structures */
typedef struct timeServerHeader {	/* Size should be multiple of DWORD */
	long int	function;
	long int	length;
	short int	version;
	short int	status;
} tsh_t;

typedef struct timeServerReply {
	tsh_t		header;
	time_t		time;
} tstr_t;
/***************************************************************************/


/**** Local Prototypes ****/

void		 ListenProcessor(int sig);

void 		 PCCleanup(pcontrol_t *pc);

pcontrol_t	*PCServerInit(char *protocol,
				void (*discon)(int),
				void *(*addrSup)(int, tbind_t *, int),
				int *error);

void		*SPXAddressSupport(int fd,
				tbind_t *tbind,
				int flags);

void 		 SPXDisconSupport(pcontrol_t *pc);

void		*TCPAddressSupport(int fd,
				tbind_t *tbind,
				int flags);

void 		 TCPDisconSupport(pcontrol_t *pc);

void		 TError(char *fmt, ...);

void		 TimeServ(pcontrol_t *pc);
/***************************************************************************/


/**** Local Variables ****/

	pcontrol_t	 *pca[PC_TABLE_SIZE];
/***************************************************************************/


/**** Main - TLI example multi-protocol server
 *
 * Input:	None
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	Exit Code				0 = Successful
 *									1 = Initialization Error
 *									2 = Protocol Error
 *									3 = Programming Error
 *
 * Comment:	Usage:  TIMESERV
 *
 ****/

void
main (void)
{
	int			 error,
				 exitControl = EXIT_NLM,
				 exitStatus = 1,
				 i;
	void		(*oldSignal)();

/* Prepare for signal processing */
	pca[I_SPX] = nilp(pcontrol_t);
	pca[I_TCP] = nilp(pcontrol_t);
	pca[I_TP4] = nilp(pcontrol_t);
	pca[I_ADSP] = nilp(pcontrol_t);
	oldSignal = signal(SIGPOLL, ListenProcessor);

/* Initialize the protocol control structures and open the protocol stacks */
	error = 0;
	pca[I_SPX] = PCServerInit("/dev/nspx", SPXDisconSupport,
			SPXAddressSupport, &error);
	pca[I_TCP] = PCServerInit("/dev/tcp", TCPDisconSupport,
			TCPAddressSupport, &error);
	/* Initialize other protocols here */

/* If any Initialization fails because of no memory then exit */
	if (error == ENOMEM)
	{
		printf("Out of Memory, unable to initialize; terminating");
		goto ErrorExit;
	}

/* Loop until escape key pressed */
	printf("Press ESCAPE key to terminate Time Server.\n");
	while (getch() != 27)
		nop;

/* Cleanup the protocols */
	exitControl = EXIT_THREAD;
	exitStatus = 0;

ErrorExit:
	signal(SIGPOLL, oldSignal);
	for (i = 0; i < PC_TABLE_SIZE; i++)
		if (pca[i])
		{
			PCCleanup(pca[i]);
			pca[i] = nilp(pcontrol_t);
		}
	ExitThread(exitControl, exitStatus);
}
/***************************************************************************/


/**** ListenProcessor - Signal handling routine to dispatch T_LISTEN events
 *
 * Input:	int			sig;		signal event
 *
 * Output:	None
 *
 * Display:	None
 *
 * Returns:	None
 *
 * Comment:	This routine will determine which protocol(s) received the
 *			connect request and create a thread to handle the connection.
 ****/

void
ListenProcessor(int sig)
{
	int			err,
				event,
				i,
				nfd;
	pcontrol_t	*localPC;
	tbind_t		*tbind;
	netbuf_t	addr;

	signal(SIGPOLL, ListenProcessor);

	for (i = 0; i < PC_TABLE_SIZE; i++)
	{
		if (pca[i]  &&  (event = t_look(pca[i]->pc_fd)))
		{
			if (event != T_LISTEN)
			{
				TError("t_look failed in signal handler, "
					   "servicing protocol %s.\n", pca[i]->pc_protocol);
				PCCleanup(pca[i]);
				pca[i] = nilp(pcontrol_t);
				continue;
			}
			if (t_listen(pca[i]->pc_fd, pca[i]->pc_tcall) == -1)
			{
				TError("t_listen failed in signal handler, "
					   "servicing protocol %s.\n", pca[i]->pc_protocol);
				t_snddis(pca[i]->pc_fd, pca[i]->pc_tcall);
				PCCleanup(pca[i]);
				pca[i] = nilp(pcontrol_t);
				continue;
			}

		/* Open a new stream to accept this connection. */
			if ((nfd = t_open(pca[i]->pc_protocol, O_RDWR,
					(tinfo_t *)0)) == -1)
			{
				TError("t_open failed in signal handler, "
					   "servicing protocol %s.\n", pca[i]->pc_protocol);
				t_snddis(pca[i]->pc_fd, pca[i]->pc_tcall);
				continue;
			}

		/* Allocate a t_bind then use the Address support routine to bind */
			if ((tbind = (tbind_t *)t_alloc(nfd, T_BIND_STR, T_ALL)) ==
					nilp(tbind_t))
			{
				printf("ENOMEM in signal handler, servicing protocol %s.\n",
						pca[i]->pc_protocol);
				t_snddis(pca[i]->pc_fd, pca[i]->pc_tcall);
				t_close(nfd);
				continue;
			}
			tbind->addr.len = tbind->addr.maxlen = pca[i]->pc_tinfo->addr;
			if (!(*pca[i]->pc_addrSup)(nfd, tbind, PS_SERVICE))
			{
				t_free((char *)tbind, T_BIND_STR);
				t_snddis(pca[i]->pc_fd, pca[i]->pc_tcall);
				t_close(nfd);
				continue;
			}

		/* Accept the connection on our new fd */
			if (t_accept(pca[i]->pc_fd, nfd, pca[i]->pc_tcall) == -1)
			{
				TError("t_accept failed in signal handler, "
					   "servicing protocol %s.\n", pca[i]->pc_protocol);
				t_free((char *)tbind, T_BIND_STR);
				t_snddis(pca[i]->pc_fd, pca[i]->pc_tcall);
				t_close(nfd);
				continue;
			}

		/* Allocate a protocol control structure and initialize it */
			if (!(localPC = (pcontrol_t *)malloc(sizeof(pcontrol_t))))
			{
				printf("ENOMEM in signal handler, servicing protocol %s.\n",
						pca[i]->pc_protocol);
				t_free((char *)tbind, T_BIND_STR);
				t_snddis(pca[i]->pc_fd, pca[i]->pc_tcall);
				t_close(nfd);
				continue;
			}
			*localPC = *pca[i];
			localPC->pc_fd = nfd;
			memmove(tbind->addr.buf, pca[i]->pc_tcall->addr.buf,
				tbind->addr.len);
			localPC->pc_tbind = tbind;

		/* Display the connection source */
			printf("Accepting connection from %s %s.\n", pca[i]->pc_protocol,
				(char *)(*pca[i]->pc_addrSup)(nfd, tbind, PS_BFORMAT));
				
			if ((err = BeginThread(TimeServ, nilp(char), STACK_SIZE,
					(int *)localPC)) == EFAILURE )
			{
				printf("Couldn't create service thread (error code %d), "
					   "servicing protocol %s.\n", errno,
					   pca[i]->pc_protocol);
				t_close(nfd);
				continue;
			}
		}
	}
}
/***************************************************************************/


/**** PCCleanup - Cleanup and Free a Protocol Control Structure
 *
 * Input:	pcontrol_t	*pc;		protocol control structure to cleanup
 *
 * Output:	None
 *
 * Display:	None
 *
 * Returns:	None
 *
 * Comment:	This routine will cleanup an allocated pcontrol_t structure
 *			by dealocating all memory.
 ****/

void
PCCleanup(pcontrol_t *pc)
{
	(*pc->pc_addrSup)(pc->pc_fd, pc->pc_tbind, PS_DEACTIVATE);	
	t_free((char *)pc->pc_tcall, T_CALL_STR);
	t_free((char *)pc->pc_tbind, T_BIND_STR);
	t_unbind(pc->pc_fd);
	t_close(pc->pc_fd);
	t_free((char *)pc->pc_tinfo, T_INFO_STR);
	free(pc);
}
/***************************************************************************/


/**** PCServerInit - Initialize a Protocol Control Structure
 *
 * Input:	char		*protocol;	   protocol name
 *			void		(*discon)();   disconnect support routine
 *			void		*(*addrSup)(); service registration routine
 *
 * Output:	int			error;		   receives error value if return NULL
 *
 * Display:	Yes
 *
 * Returns:	pcontrol_t	*pc;		   Address of initialized struct
 *
 * Comment:	This routine will allocate and initialize a protocol control
 *			structure.  If any allocation fails or the specified protocol
 *			is not available then PCServerInit cleans up before returning
 *			NULL.  Otherwise it returns the protocol control structure.
 ****/

pcontrol_t
*PCServerInit(char *protocol,
			 void (*discon)(int),
			 void *(*addrSup)(int, tbind_t *, int),
			 int *error)
{
	int			 fd;
	pcontrol_t	*pc;
	tinfo_t		 tinfo;

	if ((pc = (pcontrol_t *)malloc(sizeof(pcontrol_t))) == nilp(pcontrol_t))
	{
		*error = ENOMEM;
		goto ErrorExit0;
	}

	if ((pc->pc_fd = fd = t_open(protocol, O_RDWR, &tinfo)) == -1)
	{
		TError("t_open for transport '%s' failed", protocol);
		goto ErrorExit1;
	}
	if ((pc->pc_tinfo = (tinfo_t *)t_alloc(fd, T_INFO_STR, T_ALL)) ==
			nilp(tinfo_t))
	{
		*error = ENOMEM;
		goto ErrorExit2;
	}
	*pc->pc_tinfo = tinfo;
	if ((pc->pc_tcall = (tcall_t *)t_alloc(fd, T_CALL_STR, T_ALL)) ==
			nilp(tcall_t))
	{
		*error = ENOMEM;
		goto ErrorExit3;
	}
	ioctl(fd, I_SETSIG, S_INPUT | S_HIPRI | S_MSG);

/* Allocate & Initialize t_bind structure */
	if ((pc->pc_tbind = (tbind_t *)t_alloc(fd, T_BIND_STR, T_ALL)) ==
			nilp(tbind_t))
	{
		*error = ENOMEM;
		goto ErrorExit4;
	}
	pc->pc_tbind->addr.len = pc->pc_tbind->addr.maxlen = pc->pc_tinfo->addr;

/* Use protocol specific support routine to establish address & do t_bind */
	if (!(*addrSup)(fd, pc->pc_tbind, PS_LISTEN | PS_REGISTER | PS_ACTIVATE))
		goto ErrorExit5;

	pc->pc_protocol = protocol;
	pc->pc_discon = discon;
	pc->pc_addrSup = addrSup;

	return pc;


ErrorExit5:
	t_free((char *)pc->pc_tbind, T_BIND_STR);
ErrorExit4:
	t_free((char *)pc->pc_tcall, T_CALL_STR);
ErrorExit3:
	t_free((char *)pc->pc_tinfo, T_INFO_STR);
ErrorExit2:
	t_close(fd);
ErrorExit1:
	free(pc);
ErrorExit0:
	return nilp(pcontrol_t);
}
/***************************************************************************/


/**** SPXAddressSupport - Protocol specific address/registration support
 *
 * Input:	int 		 fd;		handle of stream disconnected
 *			tbind_t		*tbind;		bind struct to initialize
 *			int			 flags;		registration control flags
 *
 * Output:	tbind_t		*tbind;		bind struct containing address
 *
 * Display:	Yes
 *
 * Returns:	void		*ptr;		tbind, NULL or static buf
 *
 * Comment:	This function is responsible for initializing the address for
 *			specified endpoint.  It can register the service with a
 *			name service such as NetWare Directory Service or could use some
 *			other means of informing the network that the service is
 *			available and at what address.  This is currently using a very
 *			simple method.  Display the information and let the user decide
 *			what to do.
 ****/

void
*SPXAddressSupport(int fd, tbind_t *tbind, int flags)
{
	IPX_ADDR	*spxAddr;

	spxAddr = (IPX_ADDR *) (tbind->addr.buf);

/* Process format address request */
	if (flags & PS_BFORMAT)
	{
		static char	addrbuf[30];

		sprintf(addrbuf,
			"%02x%02x%02x%02x:%02x%02x%02x%02x%02x%02x;%02x%02x",
			spxAddr->ipxa_net[0] & 0xFF, spxAddr->ipxa_net[1] & 0xFF,
			spxAddr->ipxa_net[2] & 0xFF, spxAddr->ipxa_net[3] & 0xFF,
			spxAddr->ipxa_node[0] & 0xFF, spxAddr->ipxa_node[1] & 0xFF,
			spxAddr->ipxa_node[2] & 0xFF, spxAddr->ipxa_node[3] & 0xFF,
			spxAddr->ipxa_node[4] & 0xFF, spxAddr->ipxa_node[5] & 0xFF,
			spxAddr->ipxa_socket[0] & 0xFF, spxAddr->ipxa_socket[1] & 0xFF);
		return addrbuf;
	}

/* Process Address Establish requests */
	if (flags & (PS_LISTEN | PS_SERVICE))
	{
		if (flags & PS_LISTEN)
			tbind->qlen = LISTEN_Q_LENGTH;

	/* Allocate a dynamic SPX socket */
		memset(tbind->addr.buf, 0, tbind->addr.len);
		if (t_bind(fd, tbind, tbind) == -1)
		{
			TError("t_bind in SPX address support routine failed");
			return nilp(tbind_t);
		}
	}

/* Process Name service de/registration requests */
	if (flags & PS_REGISTER)
		printf("Time Server at SPX address: %s.\n",
			(char *)SPXAddressSupport(fd, tbind, PS_BFORMAT));
	if (flags & PS_DEREGISTER)
	{
		printf("Time Server removed from service for SPX clients.\n");
		return tbind;
	}

/* Process Service de/activation requests */
	if (flags & PS_ACTIVATE)
		printf("Time Server is AVAILABLE for SPX clients.\n");
	if (flags & PS_DEACTIVATE)
		printf("Time Server is UNAVAILABLE for SPX clients.\n");
	return tbind;
}
/***************************************************************************/


/**** SPXDisconSupport - determine disconnect reason code and print message
 *
 * Input:	pcontrol_t	*pc;		control structure
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	Determine why the connection was terminated, by checking the
 *			reason code returned from t_rcvdis.  These are SPX specific codes
 *			and a message is printed accordingly.
 ****/

void
SPXDisconSupport(pcontrol_t *pc)
{
	tdiscon_t	 discon;

	if (t_rcvdis(pc->pc_fd, &discon) == -1)
	{
		t_error("t_rcvdis failed");
		ExitThread(0, 2);
	}
	switch(discon.reason)
	{
	case TLI_SPX_CONNECTION_FAILED:
		printf("Connection with %s %s terminated because\n"
			"\tthe connection failed.\n", pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;

	case TLI_SPX_CONNECTION_TERMINATED:
		printf("Connection with %s %s terminated by client.\n",
			pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;

	case TLI_SPX_MALFORMED_PACKET:
		printf("Connection with %s %s terminated because\n"
			"\tof internal SPX error.\n", pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;

	default:
		printf("Connection with %s %s terminated for unknown reason.\n",
			pc->pc_protocol,
			SPXAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
	}
}
/***************************************************************************/


/**** TCPAddressSupport - Protocol specific address/registration support
 *
 * Input:	int 		 fd;		handle of stream disconnected
 *			tbind_t		*tbind;		bind struct to initialize
 *			int			 flags;		registration control flags
 *
 * Output:	tbind_t		*tbind;		bind struct containing address
 *
 * Display:	Yes
 *
 * Returns:	void		*ptr;		tbind, NULL or static buf
 *
 * Comment:	This function is responsible for initializing the address for
 *			specified endpoint.  It can register the service with a
 *			nameservice such as NetWare Directory Service or could use some
 *			other means of informing the network that the service is
 *			available and at what address.  This is currently using a very
 *			simple method.  Display the information and let the user decide
 *			what to do.
 ****/

void
*TCPAddressSupport(int fd, tbind_t *tbind, int flags)
{
	ipaddr_t	*ipAddr;

	ipAddr = (ipaddr_t *)tbind->addr.buf;

/* Process format address request */
	if (flags & PS_BFORMAT)
	{
		static char	addrbuf[30];

		sprintf(addrbuf, "%d.%d.%d.%d, port %x",
			((char *)&ipAddr->sin_addr.s_addr)[0],
			((char *)&ipAddr->sin_addr.s_addr)[1],
			((char *)&ipAddr->sin_addr.s_addr)[2],
			((char *)&ipAddr->sin_addr.s_addr)[3],
			htons(ipAddr->sin_port));
		return addrbuf;
	}

/* Process Address Establish requests */
	if (flags & (PS_LISTEN | PS_SERVICE))
	{
		if (flags & PS_LISTEN)
			tbind->qlen = LISTEN_Q_LENGTH;

	/* Initialize everything to 0, then set required fields */
		memset(ipAddr, 0, tbind->addr.len);
		ipAddr->sin_family = AF_INET;

	/* Allocate a dynamic TCP port */
		if (t_bind(fd, tbind, tbind) == -1)
		{
			TError("t_bind in TCP address support routine failed");
			return FALSE;
		}
	}

/* Process Name service de/registration requests */
	if (flags & PS_REGISTER)
		printf("Time Server at TCP address:  %s.\n",
			(char *)TCPAddressSupport(fd, tbind, PS_BFORMAT));
	if (flags & PS_DEREGISTER)
	{
		printf("Time Server removed from service for TCP clients.\n");
		return tbind;
	}

/* Process Service de/activation requests */
	if (flags & PS_ACTIVATE)
		printf("Time Server is AVAILABLE for TCP clients.\n");

	if (flags & PS_DEACTIVATE)
		printf("Time Server is UNAVAILABLE for TCP clients.\n");

	return tbind;
}
/***************************************************************************/


/**** TCPDisconSupport - determine disconnect reason code and print message
 *
 * Input:	pcontrol_t	*pc;		control structure
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	Determine why the connection was terminated, by checking the
 *			reason code returned from t_rcvdis.  These are TCP specific codes
 *			and a message is printed accordingly.
 ****/

void
TCPDisconSupport(pcontrol_t *pc)
{
	tdiscon_t	 discon;

	if (t_rcvdis(pc->pc_fd, &discon) == -1)
	{
		t_error("t_rcvdis failed");
		ExitThread(0, 2);
	}
	switch(discon.reason)
	{
	case ETIMEDOUT:
		printf("Connection with %s %s terminated because\n"
			"\tthe connection failed.\n", pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;

	case ECONNRESET:
		printf("Connection with %s %s terminated by client.\n",
			pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;

	case ENETDOWN:
		printf("Connection with %s %s terminated because\n"
			"\tTCP is either unloading or has been unloaded.\n",
			pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;

	default:
		printf("Connection with %s %s terminated for unknown reason.\n",
			pc->pc_protocol,
			TCPAddressSupport(pc->pc_fd, pc->pc_tbind, PS_BFORMAT));
		break;
	}
}
/***************************************************************************/


/**** TError - t_error with replacable parameters
 *
 * Input:	char		*fmt;		display string
 *
 * Output:	None
 *
 * Display:	Yes
 *
 * Returns:	None
 *
 * Comment:	t_error does not provide for replacable parameters.  This
 *			routine front ends t_error to provide this functionality.
 *			It is susceptable to local buffer overflow problems.
 *
 ****/

void
TError(char *fmt, ...)
{
	va_list			argptr;
	char			buf[512];

	va_start(argptr, fmt);
	vsprintf(buf, fmt, argptr);
	va_end(argptr);
	t_error(buf);
}
/***************************************************************************/


/**** TimeServ - Send Receive Loop for new thread
 *
 * Input:	pcontrol_t	*pc;		control struct for this connection
 *
 * Output:	None
 *
 * Display:	None
 *
 * Returns:	None
 *
 * Comment:	This is the send/receive thread for each connection.  It is
 *			started by the BeginThread call in ListenProcessor.  If any
 *			error occurs on any TLI call it terminates the thread.
 *			NOTE:  The parameter "pc" is released.
 ****/

void
TimeServ(pcontrol_t *pc)
{
	char		 buf[100];
	int			 flags,
				 replyLength,
				 nfd;
	
	nfd = pc->pc_fd;
	flags = 0;			/* Initialize for t_rcv call */

/* Loop until t_rcv fails when client closes down the connection. */
	while (t_rcv(nfd, buf, sizeof(buf), &flags) != -1)
	{
		if (((tsh_t *)buf)->function == TS_TIME_REQUEST)
		{
			((tsh_t *)buf)->status = TS_REQUEST_COMPLETE;
			((tstr_t *)buf)->time = time(nilp(time_t));
			replyLength = sizeof(tstr_t);
		}
		else
		{
			((tsh_t *)buf)->status = TS_INVALID_REQUEST;
			replyLength = sizeof(tsh_t);
		}

	/* Send Reply back to the Client */
		((tsh_t *)buf)->length = replyLength;
		if (t_snd(nfd, buf, sizeof(tstr_t), 0) == -1)
		{
			TError("t_snd failed in connection handler.\n");
			t_close(nfd);
			ExitThread(0, 1);
		}
	}
	/* break out of receive mode and find out why */

	/* Finish the disconnect if the client started one. */
	if (t_errno == TLOOK  &&  t_look(nfd) == T_DISCONNECT)
		(*pc->pc_discon)(pc);
	else	/* some other error so disconnect */
	{
		TError("t_rcv failed in connection handler; t_look error is %lx",
			t_look(nfd));
		if (t_snddis(nfd, nilp(tcall_t)) == -1)
			TError("t_snddis failed in connection handler.\n");
	}
	/* Close the transport stream. */
 	t_unbind(nfd);
	t_close(nfd);
	t_free((char *)pc->pc_tbind, T_BIND_STR);
	free((char *)pc);
	ExitThread(EXIT_THREAD, 0);
}
/***************************************************************************/
/***************************************************************************/

